//This code runs experiments in C language.
//The code is a modification of exemplary code from COCO (http://coco.gforge.inria.fr/) aimed at running CMA-ES (http://cma.gforge.inria.fr/cmaesintro.html).
//modified by Rafał Biedrzycki


#include <stdio.h>
#include <stdbool.h>
#include <string.h>

#include "cmaes_interface.h"
#include "boundary_transformation.h"
#include "coco.h"

static int function_id = 0;
static coco_random_state_t *randState=NULL;



void initial_solution_proposal(coco_problem_t* problem, double initial_solution[], int restart_number){
	const size_t number_of_variables = coco_problem_get_dimension(problem);
	const double *lower = coco_problem_get_smallest_values_of_interest(problem);
	const double *upper = coco_problem_get_largest_values_of_interest(problem);
	if( restart_number <=  0 || coco_problem_get_number_of_constraints(problem) > 0){
	  coco_problem_get_initial_solution(problem, initial_solution);
	  return;
	}
	for(int i=0; i<number_of_variables; ++i){
	  lower[i] + (upper[i]-lower[i]) *(coco_random_uniform(randState)+coco_random_uniform(randState))/2;
	}
}

void cma_optimizer(coco_problem_t *problem) {
  
  double *const *X, *y;

  /* Extract number of variables and bounds from problem */
  const size_t number_of_variables = coco_problem_get_dimension(problem);
  const double *lower = coco_problem_get_smallest_values_of_interest(problem);
  const double *upper = coco_problem_get_largest_values_of_interest(problem);
  double* x_in_boundsd = cmaes_NewDouble(number_of_variables);

  /* Allocate vectors for initial solution and initial sigma */
  double *initial_solution = coco_allocate_vector(number_of_variables);
  coco_problem_get_initial_solution(problem, initial_solution);


  double *initial_sigma = coco_allocate_vector(number_of_variables);
  cmaes_boundary_transformation_t boundaries;

  cmaes_boundary_transformation_init(&boundaries, lower, upper, number_of_variables);
  const char *stop;
  for (size_t i = 0; i < number_of_variables; ++i) {
    initial_sigma[i] = 0.3* (upper[i] - lower[i]);
  }
  
  int initial_budget = 1e7 * number_of_variables;
  while(  coco_problem_get_evaluations(problem)  < initial_budget && ! coco_problem_final_target_hit( problem)){
    cmaes_t cma;
    int budget_left = initial_budget - coco_problem_get_evaluations(problem);

    /* Initialize CMA-ES without parameter file */ 
    cmaes_init_para(&cma, number_of_variables, initial_solution, initial_sigma, 0, 0, "no");
    cma.sp.stopMaxFunEvals = budget_left;
    cma.sp.filename = strdup("no");
    y = cmaes_init_final(&cma);

    /* Main evolutionary loop */
    while (!(stop=cmaes_TestForTermination(&cma))) {
      X = cmaes_SamplePopulation(&cma);

      for (size_t i = 0; i < cmaes_Get(&cma, "lambda"); ++i) {
        cmaes_boundary_transformation(&boundaries, X[i], x_in_boundsd, number_of_variables);
        coco_evaluate_function(problem, x_in_boundsd, &y[i]);
      }
      cmaes_UpdateDistribution(&cma, y);
    }
    puts(stop);

    /* Cleanup and free memory */
    cmaes_exit(&cma);
    initial_solution_proposal(problem, initial_solution, 77);
  }
  coco_free_memory(initial_solution);
  coco_free_memory(initial_sigma);
}

/*
 * @param suite_options A string of pairs "key: value" used to filter the suite (especially useful for
 * parallelizing the experiments). Supported options:
 * - "dimensions: LIST", where LIST is the list of dimensions to keep in the suite (range-style syntax is
 * not allowed here), poprzecinku: 2,3,5
 * - "dimension_indices: VALUES", where VALUES is a list or a range of dimension indices (starting from 1) to keep
 * in the suite, and
 * - "function_indices: VALUES", where VALUES is a list or a range of function indices (starting from 1) to keep
 * in the suite, and
 * - "instance_indices: VALUES", where VALUES is a list or a range of instance indices (starting from 1) to keep
 * in the suite.
*/
int main() {
  char algorithm_id[200] = "C-CMA";
  randState= coco_random_new(1234);

  coco_suite_t *suite;
  coco_observer_t *observer;
  coco_problem_t *problem;
  char *observer_options;

  observer_options = coco_strdupf("result_folder: %s "
                   "algorithm_name: %s "
                   "algorithm_info: \"C CMA-ES\"", algorithm_id, algorithm_id);

  suite = coco_suite("bbob", "year: 2009", NULL);
  observer = coco_observer("bbob", observer_options);

  while ((problem = coco_suite_get_next_problem(suite, observer)) != NULL) {
    cma_optimizer(problem);
  }

  coco_observer_free(observer);
  coco_suite_free(suite);
  coco_free_memory(observer_options);
  coco_random_free(randState);

  return 0;
}

