//This code runs experiments in Java language.
//The code is a modification of exemplary code from COCO (http://coco.gforge.inria.fr/) aimed at running CMA-ES (http://cma.gforge.inria.fr/cmaesintro.html).
//modified by Rafał Biedrzycki

import java.util.Random;
import java.util.Locale;
import fr.inria.optimization.cmaes.CMAEvolutionStrategy;
import fr.inria.optimization.cmaes.fitness.IObjectiveFunction;


class FitFunWrapper implements IObjectiveFunction{ 
  Problem problem_;
  double[] lowerBounds_;
  double[] upperBounds_;

  public FitFunWrapper( Problem problem, double[] lowerBounds, double[] upperBounds ){
    problem_ = problem;
    lowerBounds_ = lowerBounds;
    upperBounds_ = upperBounds;
  }

  public double valueOf (double[] x) {
    return problem_.evaluateFunction( x )[0];
  }
  
  public boolean isFeasible(double[] x) {
    for( int i=0; i<x.length; ++i){
      if( x[i]>upperBounds_[i] || x[i] <lowerBounds_[i] ){
        return false;
      }
    }
    return true; 
  } 
};


/**
 * An example of benchmarking random search on a COCO suite. 
 *
 * Set the parameter BUDGET_MULTIPLIER to suit your needs.
 */
public class ExampleExperiment {

	/**
	 * The maximal budget for evaluations done by an optimization algorithm equals 
	 * dimension * BUDGET_MULTIPLIER.
	 * Increase the budget multiplier value gradually to see how it affects the runtime.
	 */
	public static final int BUDGET_MULTIPLIER = 10000000;
	
	/**
	 * The maximal number of independent restarts allowed for an algorithm that restarts itself. 
	 */
	public static final int INDEPENDENT_RESTARTS = 10000;
	
	/**
	 * The random seed. Change if needed.
	 */
	public static final long RANDOM_SEED = 0xdeadbeef;

	/**
	 * The main method initializes the random number generator and calls the example experiment on the
	 * bi-objective suite.
	 */
	public static void main(String[] args) {
		Locale.setDefault(new Locale("en", "US", "LIN"));
		
		Random randomGenerator = new Random(RANDOM_SEED);

		/* Change the log level to "warning" to get less output */
		CocoJNI.cocoSetLogLevel("warning");

		System.out.println("Running the example experiment... (might take time, be patient)");
		System.out.flush();
		     
		/* Start the actual experiments on a test suite and use a matching logger, for
		* example one of the following:
		*
		*   bbob                 24 unconstrained noiseless single-objective functions
		*   bbob-biobj           55 unconstrained noiseless bi-objective functions
		*   bbob-biobj-ext       92 unconstrained noiseless bi-objective functions
		*   bbob-largescale      24 unconstrained noiseless single-objective functions in large dimension
		*   bbob-constrained     48 constrained noiseless single-objective functions
		*
		* Adapt to your need. Note that the experiment is run according
		* to the settings, defined in exampleExperiment(...) below.
		*/
		exampleExperiment("bbob", "bbob", randomGenerator);

		System.out.println("Done!");
		System.out.flush();

		return;
	}
	
	public static void exampleExperiment(String suiteName, String observerName, Random randomGenerator) {

		try {

			/* Set some options for the observer. See documentation for other options. */
			final String observerOptions = 
					  "result_folder: JavaCma" + suiteName + " " 
					+ "algorithm_name: "+suiteName + " " 
					+ "algorithm_info: \"Java CMA-ES\"";

			/* Initialize the suite and observer.
			* For more details on how to change the default options, see
			* http://numbbo.github.io/coco-doc/C/#suite-parameters and
			* http://numbbo.github.io/coco-doc/C/#observer-parameters. */
			Suite suite = new Suite(suiteName, "year: 2009", "");
			Observer observer = new Observer(observerName, observerOptions);
			Benchmark benchmark = new Benchmark(suite, observer);

			/* Initialize timing */
			Timing timing = new Timing();
			
			/* Iterate over all problems in the suite */
      			Problem problem;
			while ((problem = benchmark.getNextProblem()) != null) {

				int dimension = problem.getDimension();

				runCMAES(problem,
						       dimension,
						       problem.getNumberOfObjectives(),
						       problem.getNumberOfConstraints(),
						       problem.getSmallestValuesOfInterest(),
						       problem.getLargestValuesOfInterest(),
						       (long) (dimension * BUDGET_MULTIPLIER),
						       randomGenerator);


				/* Keep track of time */
				timing.timeProblem(problem);
			}

			/* Output the timing data */
			timing.output();

			benchmark.finalizeBenchmark();

		} catch (Exception e) {
			System.err.println(e.toString());
		}
	}

	static double [] projection( double[] indiv, double min, double max ){
		for(int i = 0; i < indiv.length; ++i) {
			if( indiv[i]<min )
				indiv[i]=min;
			else if(indiv[i]>max)
				indiv[i]=max;
		}
		return indiv;
	}

	/** 
	 * A simple random search algorithm that can be used for single- as well as multi-objective 
	 * optimization.
	 */
	public static void runCMAES( Problem problem,
		                          int dimension, 
		                          int numberOfObjectives,
					   int numberOfConstraints,
		                          double[] lowerBounds,
		                          double[] upperBounds, 
		                          long maxBudget, 
		                          Random randomGenerator) {
		FitFunWrapper fitfun = new FitFunWrapper( problem, lowerBounds, upperBounds );
		double SIGMA = 0.3*(upperBounds[0]-lowerBounds[0]);


		double[] startingSol=new double[dimension]; 
		for(int j=0; j< dimension; ++j){
			startingSol[j]= lowerBounds[j] + (upperBounds[j]-lowerBounds[j])/2;
		} 

		while( problem.getEvaluations()<maxBudget && !problem.isFinalTargetHit() ){
			CMAEvolutionStrategy cma = new CMAEvolutionStrategy();
			cma.readProperties(); // read options, see file CMAEvolutionStrategy.properties
			cma.setDimension( dimension ); // overwrite some loaded properties
			cma.setInitialStandardDeviation( SIGMA ); // also a mandatory setting 
			cma.setInitialX( startingSol ); // in each dimension, also setTypicalX can be used
			cma.options.stopMaxFunEvals= maxBudget - problem.getEvaluations();

			double[] fitness = cma.init();  // new double[cma.parameters.getPopulationSize()];

			while(cma.stopConditions.getNumber() == 0) {
				double[][] pop = cma.samplePopulation(); // get a new population of solutions
				for(int i = 0; i < pop.length; ++i) {    // for each candidate solution i
					int numberOfTrials=0;
					while (!fitfun.isFeasible(pop[i]) && numberOfTrials<100 ){     //   not located on (or very close to) the domain boundary,  
						pop[i] = cma.resampleSingle(i);    //   initialX is feasible and initialStandardDeviations are  
						++numberOfTrials;
					}
					if( numberOfTrials==100 && !fitfun.isFeasible(pop[i]) ){
						pop[i] = projection( pop[i], lowerBounds[0], upperBounds[0] ); 
					}
					fitness[i] = fitfun.valueOf(pop[i]); // fitfun.valueOf() is to be minimized
				}//for
				cma.updateDistribution(fitness);         // pass fitness array to update search distribution
			}//while
			// evaluate mean value as it is the best estimator for the optimum
			cma.setFitnessOfMeanX(fitfun.valueOf(cma.getMeanX())); // updates the best ever solution 

			cma.println("Terminated due to");
			for (String s : cma.stopConditions.getMessages())
				cma.println("  " + s);
			cma.println("best function value " + cma.getBestFunctionValue() 
				  + " at evaluation " + cma.getBestEvaluationNumber());

			for(int j=0; j< dimension; ++j){
				startingSol[j]= lowerBounds[j] + (upperBounds[j]-lowerBounds[j]) *(randomGenerator.nextDouble()+randomGenerator.nextDouble())/2.0;

			} 
		}
	}
}

